from typing import Dict, Union, List

SchemaType = Union[str, List, Dict]

FINGERPRINT_ALGORITHMS: str

def load_schema(schema_path: str) -> Dict: ...
def parse_schema(
    schema: SchemaType, expand: bool, _write_hint: bool, _force: bool
) -> SchemaType: ...
def fullname(schema: Dict) -> str: ...
def expand_schema(schema: Dict) -> Dict: ...
def load_schema_ordered(ordered_schemas: List[str]) -> Dict: ...
def to_parsing_canonical_form(schema: SchemaType) -> str: ...
def fingerprint(canonical_form: str, algorithm: str) -> str: ...
